/*
 * Copyright (c) 2012 Michael Cardell Widerkrantz <mc@hack.org> 
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.  
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>

#include <openssl/ssl.h>
#include <openssl/pem.h>
#include <openssl/evp.h>
#include <openssl/sha.h>

static struct buffer *
	buffer_alloc(u_int len);
static void
	buffer_free(struct buffer *buffer);
static void
	put_u32(void *vp, u_int32_t v);
static int
	buffer_int(struct buffer *buffer, u_int value);
static int
	buffer_string(struct buffer *buffer, const void *buf, u_int len);
static int
	buffer_bignum(struct buffer *buffer, BIGNUM *bn);
static int
	fingerprint(EVP_PKEY *key, char *mdstr);

struct buffer
{
	u_char  *buf;   /* Data. */
	u_int    len;   /* Length of data in bytes. */
	u_int    size;  /* Allocated size of buf. */
};

static struct buffer *
buffer_alloc(u_int len)
{
	struct buffer	*buffer;
	
	if (NULL == (buffer = calloc(1, sizeof (struct buffer)))) {
		return NULL;
	}
	buffer->len = 0;
	buffer->size = len;
	if (NULL == (buffer->buf = malloc(len))) {
		free(buffer);
		return NULL;
	}

	return buffer;
}

static void
buffer_free(struct buffer *buffer)
{
	if (buffer == NULL)
		return;
	if (buffer->buf != NULL)
		free(buffer->buf);
	free(buffer);
}

static void
put_u32(void *vp, u_int32_t v)
{
	u_char *p = (u_char *)vp;

	p[0] = (u_char)(v >> 24) & 0xff;
	p[1] = (u_char)(v >> 16) & 0xff;
	p[2] = (u_char)(v >> 8) & 0xff;
	p[3] = (u_char)v & 0xff;
}

static int
buffer_int(struct buffer *buffer, u_int value)
{
	char buf[4];

        if (buffer->len + 4 > buffer->size)
            return (-1);
        
	put_u32(buf, value);
	memcpy(&buffer->buf[buffer->len], buf, 4);
	buffer->len += 4;

        return (0);
}

static int
buffer_string(struct buffer *buffer, const void *buf, u_int len)
{
        if (buffer->len + len > buffer->size)
            return (-1);
    
	memcpy(&buffer->buf[buffer->len], buf, len);
	buffer->len += len;

        return (0);
}

static int
buffer_bignum(struct buffer *buffer, BIGNUM *bn)
{
	unsigned char	*buf;	
	u_int		 len;
	int		 oi;
	
	len = BN_num_bytes(bn);
        if (buffer->len + len > buffer->size) {
            return (-1);
        }

	if (NULL == (buf = malloc(len))) {
		return (-1);
	}
	oi = BN_bn2bin(bn, buf);
	if (oi < 0 || (u_int)oi != len) {
		fprintf(stderr, " BN_bn2bin() failed: "
		      "oi %d != bin_size %d\n", oi, len);
		free(buf);
		return (-1);
	}
	buffer_string(buffer, buf, len);
	free(buf);

	return (0);
}

static int
fingerprint(EVP_PKEY *key, char *mdstr)
{
	RSA		*rsa = NULL;
	struct buffer	*buffer = NULL;
	unsigned char	*md;        
	u_int32_t	 elen, nlen;
	u_int		 bytes;
	int		 ret;
	
	rsa = EVP_PKEY_get1_RSA(key);
	elen = BN_num_bytes(rsa->e);
	nlen = BN_num_bytes(rsa->n);

	/* Allocate space for two bignums and their length. */
	bytes = elen + nlen + 2 * 4;
	if (NULL == (buffer = buffer_alloc(bytes))) {
		fprintf(stderr, "malloc: out of memory");
		return (-1);
	}

	/* RSA exponent e. First length, then e itself. */
	if (buffer_int(buffer, elen) != 0) {
            printf("Out of space.\n");
	    ret = -1;
            goto done;
        }
	if (buffer_bignum(buffer, rsa->e) != 0) {
            printf("Out of space.\n");
	    ret = -1;
	    goto done;
        }
        
	/* RSA modulus n. First length, then n itself. */
	if (buffer_int(buffer, nlen) != 0) {
            printf("Out of space.\n");
	    ret = -1;	    
            goto done;
        }
	if (buffer_bignum(buffer, rsa->n) != 0) {
            printf("Out of space.\n");
	    ret = -1;	    
            goto done;
        }

        /* SHA-1 digest of both bignums and their length. */
	md = SHA1(buffer->buf, buffer->len, NULL);
	snprintf(mdstr, 41, "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x"
	    "%02x%02x%02x%02x%02x%02x%02x%02x", md[0], md[1], md[2],
	    md[3], md[4], md[5], md[6], md[7], md[8], md[9], md[10],
	    md[11], md[12], md[13], md[14], md[15], md[16], md[17],
            md[18], md[19]);

	ret = 0;
done:
	if (buffer != NULL)
		buffer_free(buffer);
	if (rsa != NULL)
		free(rsa);

	return ret;
}

int
readfile(char *file)
{
	EVP_PKEY *localkey = NULL;
	FILE *fp;
        char mdstr[41];
	
	if ((fp = fopen(file, "r")) == NULL) {
		perror("fopen");
		exit(-1);
	}

	localkey = PEM_read_PUBKEY(fp, NULL, NULL, NULL);
	fclose(fp);

	if (localkey == NULL) {
		fprintf(stderr, "couldn't parse key\n");
		exit(-1);
	}

        fingerprint(localkey, mdstr);
	printf("%s:\n%s\n", file, mdstr);

	return (0);
}

int
main(int argc, char *argv[])
{
	if (argc != 2) {
		fprintf(stderr, "rsafp <file>\n");
		exit(1);
	}
	readfile(argv[1]);
	exit(0);
}
